import torch
from torch.distributions import (
    Distribution,
    Normal,
)
from torch.nn.functional import one_hot
import numpy as np
from typing import (
    Literal,
    Callable,
    Dict,
    List,
    Tuple,
)
from .base import (
    EmbeddingModuleMixin,
    BaseModuleClass,
    LossOutput,
)


class CVAE(EmbeddingModuleMixin, BaseModuleClass):
    """
    Conditional variational auto-encoder for scRNA-seq data. 

    Adapted from https://github.com/scverse/scvi-tools. 

    Parameters
    ----------
    n_input: int 
        Number of input features.
    n_batch: int, default 0  
        Number of batches. If 0, no batch correction is performed.
    n_labels: int, default 0
        Number of labels.
    n_hidden: int, default 128
        Number of nodes per hidden layer. Passed into ``scvi.nn.Encoder`` and ``scvi.nn.DecoderSCVI``.
    n_latent: int, default 10 
        Dimensionality of the latent space.
    n_layers: int, default 1 
        Number of hidden layers. Passed into ``scvi.nn.Encoder`` and ``scvi.nn.DecoderSCVI``.
    n_continuous_cov: int, default 0 
        Number of continuous covariates.
    n_cats_per_cov: list of int, optional, default None 
        A list of integers containing the number of categories for each categorical covariate.
    dropout_rate: float, default 0.1 
        Dropout rate. Passed into ``scvi.nn.Encoder`` but not ``scvi.nn.DecoderSCVI``.
    dispersion: {"gene", "gene-batch", "gene-label", "gene-cell"}, default "gene"
        Flexibility of the dispersion parameter when ``gene_likelihood`` is either "nb" or
        "zinb". One of the following:
        * "gene": parameter is constant per gene across cells.
        * "gene-batch": parameter is constant per gene per batch.
        * "gene-label": parameter is constant per gene per label.
        * "gene-cell": parameter is constant per gene per cell.
    log_variational: bool, default True
        If True, use ``torch.log1p`` on input data before encoding for numerical stability
        (not normalization).
    gene_likelihood: {"zinb", "nb", "poisson"}, default "zinb"
        Distribution to use for reconstruction in the generative process. One of the following:
        * "nb": ``scvi.distributions.NegativeBinomial``.
        * "zinb": ``scvi.distributions.ZeroInflatedNegativeBinomial``.
        * "poisson": ``scvi.distributions.Poisson``.
    latent_distribution: {"normal", "ln"}, default "normal"
        Distribution to use for the latent space. One of the following:
        * "normal": isotropic normal.
        * "ln": logistic normal with normal params N(0, 1).
    encode_covariates: bool, default False
        If True, covariates are concatenated to gene expression prior to passing through
        the encoder(s). Else, only gene expression is used.
    deeply_inject_covariates: bool, default True
        If True and ``n_layers > 1``, covariates are concatenated to the outputs of hidden
        layers in the encoder(s) (if ``encoder_covariates`` is True) and the decoder prior to
        passing through the next layer.
    batch_representation: {"one-hot", "embedding"}, default "one-hot"
        ``EXPERIMENTAL`` Method for encoding batch information. One of the following:
        * "one-hot": represent batches with one-hot encodings.
        * "embedding": represent batches with continuously-valued embeddings using ``scvi.nn.Embedding``.
        Note that batch representations are only passed into the encoder(s) if
        ``encode_covariates`` is True.
    use_batch_norm: {"encoder", "decoder", "none", "both"}, default "both"
        Specifies where to use ``torch.nn.BatchNorm1d`` in the model. One of the following:
        * "none": don't use batch norm in either encoder(s) or decoder.
        * "encoder": use batch norm only in the encoder(s).
        * "decoder": use batch norm only in the decoder.
        * "both": use batch norm in both encoder(s) and decoder.
        Note: if ``use_layer_norm`` is also specified, both will be applied (first
        ``torch.nn.BatchNorm1d``, then ``torch.nn.LayerNorm``).
    use_layer_norm: {"encoder", "decoder", "none", "both"}, default "none"
        Specifies where to use ``torch.nn.LayerNorm`` in the model. One of the following:
        * "none": don't use layer norm in either encoder(s) or decoder.
        * "encoder": use layer norm only in the encoder(s).
        * "decoder": use layer norm only in the decoder.
        * "both": use layer norm in both encoder(s) and decoder.
        Note: if ``use_batch_norm`` is also specified, both will be applied (first
        ``torch.nn.BatchNorm1d``, then ``torch.nn.LayerNorm``).
    use_size_factor_key: bool, default False
        If True, use the ``anndata.AnnData.obs`` column as defined by the
        ``size_factor_key`` parameter in the model's ``setup_anndata`` method as the scaling
        factor in the mean of the conditional distribution. Takes priority over
        ``use_observed_lib_size``.
    use_observed_lib_size: bool, default True
        If True, use the observed library size for RNA as the scaling factor in the mean of the
        conditional distribution.
    library_log_means: np.ndarray, optional, default None
        ``numpy.ndarray`` of shape (1, n_batch) of means of the log library sizes that
        parameterize the prior on library size if ``use_size_factor_key`` is False and
        ``use_observed_lib_size`` is False.
    library_log_vars: np.ndarray, optional, default None
        ``numpy.ndarray`` of shape (1, n_batch) of variances of the log library sizes
        that parameterize the prior on library size if ``use_size_factor_key`` is False and
        ``use_observed_lib_size`` is False.
    var_activation: Callable, default None
        Callable used to ensure positivity of the variance of the variational distribution. Passed
        into ``scvi.nn.Encoder``. Defaults to ``torch.exp``.
    extra_encoder_kwargs: dict, optional, default None
        Additional keyword arguments passed into ``scvi.nn.Encoder``.
    extra_decoder_kwargs: dict, optional, default None
        Additional keyword arguments passed into ``scvi.nn.DecoderSCVI``.
    batch_embedding_kwargs: dict, optional, default None
        Keyword arguments passed into ``scvi.nn.Embedding`` if ``batch_representation`` is
        set to "embedding".
    adaptive_library: bool, default True
        If True, the model will adaptively decode the library size based on the conditional information 
        instead of using batch information only. Note that this is more flexible because it can take 
        both batch information and other conditional information into account. This is typically used
        when conditional information is provided by an external model.
    """

    def __init__(
        self,
        n_input: int,
        n_batch: int = 0,
        n_labels: int = 0,
        n_hidden: int = 128,
        n_latent: int = 10,
        n_layers: int = 1,
        n_continuous_cov: int = 0,
        n_cats_per_cov: List[int] | None = None,
        dropout_rate: float = 0.1,
        dispersion: Literal["gene", "gene-batch", "gene-label",
                            "gene-cell"] = "gene",
        log_variational: bool = True,
        gene_likelihood: Literal["zinb", "nb", "poisson"] = "zinb",
        latent_distribution: Literal["normal", "ln"] = "normal",
        encode_covariates: bool = False,
        deeply_inject_covariates: bool = True,
        batch_representation: Literal["one-hot", "embedding"] = "one-hot",
        use_batch_norm: Literal["encoder", "decoder", "none", "both"] = "both",
        use_layer_norm: Literal["encoder", "decoder", "none", "both"] = "none",
        use_size_factor_key: bool = False,
        use_observed_lib_size: bool = True,
        library_log_means: np.ndarray | None = None,
        library_log_vars: np.ndarray | None = None,
        var_activation: Callable[[torch.Tensor], torch.Tensor] = None,
        extra_encoder_kwargs: dict | None = None,
        extra_decoder_kwargs: dict | None = None,
        batch_embedding_kwargs: dict | None = None,
        adaptive_library: bool = True,
    ) -> "CVAE":
        from scvi.nn import DecoderSCVI, Encoder

        super().__init__()

        self.dispersion = dispersion
        self.n_latent = n_latent
        self.log_variational = log_variational
        self.gene_likelihood = gene_likelihood
        self.n_batch = n_batch
        self.n_labels = n_labels
        self.latent_distribution = latent_distribution
        self.encode_covariates = encode_covariates
        self.use_size_factor_key = use_size_factor_key
        self.use_observed_lib_size = use_size_factor_key or use_observed_lib_size

        if not self.use_observed_lib_size:
            if library_log_means is None or library_log_vars is None:
                raise ValueError(
                    "If not using observed_lib_size, "
                    "must provide library_log_means and library_log_vars.")

            self.register_buffer("library_log_means",
                                 torch.from_numpy(library_log_means).float())
            self.register_buffer("library_log_vars",
                                 torch.from_numpy(library_log_vars).float())

        if self.dispersion == "gene":
            self.px_r = torch.nn.Parameter(torch.randn(n_input))
        elif self.dispersion == "gene-batch":
            self.px_r = torch.nn.Parameter(torch.randn(n_input, n_batch))
        elif self.dispersion == "gene-label":
            self.px_r = torch.nn.Parameter(torch.randn(n_input, n_labels))
        elif self.dispersion == "gene-cell":
            pass
        else:
            raise ValueError(
                "`dispersion` must be one of 'gene', 'gene-batch', 'gene-label', 'gene-cell'."
            )

        self.batch_representation = batch_representation
        if self.batch_representation == "embedding":
            self.init_embedding('batch_ids', n_batch,
                                **(batch_embedding_kwargs or {}))
            batch_dim = self.get_embedding('batch_ids').embedding_dim
        elif self.batch_representation != "one-hot":
            raise ValueError(
                "`batch_representation` must be one of 'one-hot', 'embedding'.")

        use_batch_norm_encoder = use_batch_norm == "encoder" or use_batch_norm == "both"
        use_batch_norm_decoder = use_batch_norm == "decoder" or use_batch_norm == "both"
        use_layer_norm_encoder = use_layer_norm == "encoder" or use_layer_norm == "both"
        use_layer_norm_decoder = use_layer_norm == "decoder" or use_layer_norm == "both"

        n_input_encoder = n_input + n_continuous_cov * encode_covariates
        if self.batch_representation == "embedding":
            n_input_encoder += batch_dim * encode_covariates
            cat_list = list([] if n_cats_per_cov is None else n_cats_per_cov)
        else:
            cat_list = [
                n_batch
            ] + list([] if n_cats_per_cov is None else n_cats_per_cov)

        encoder_cat_list = cat_list if encode_covariates else None
        _extra_encoder_kwargs = extra_encoder_kwargs or {}
        self.z_encoder = Encoder(
            n_input_encoder,
            n_latent,
            n_cat_list=encoder_cat_list,
            n_layers=n_layers,
            n_hidden=n_hidden,
            dropout_rate=dropout_rate,
            distribution=latent_distribution,
            inject_covariates=deeply_inject_covariates,
            use_batch_norm=use_batch_norm_encoder,
            use_layer_norm=use_layer_norm_encoder,
            var_activation=var_activation,
            return_dist=True,
            **_extra_encoder_kwargs,
        )
        # l encoder goes from n_input-dimensional data to 1-d library size
        self.l_encoder = Encoder(
            n_input_encoder,
            1,
            n_layers=1,
            n_cat_list=encoder_cat_list,
            n_hidden=n_hidden,
            dropout_rate=dropout_rate,
            inject_covariates=deeply_inject_covariates,
            use_batch_norm=use_batch_norm_encoder,
            use_layer_norm=use_layer_norm_encoder,
            var_activation=var_activation,
            return_dist=True,
            **_extra_encoder_kwargs,
        )
        n_input_decoder = n_latent + n_continuous_cov
        if self.batch_representation == "embedding":
            n_input_decoder += batch_dim

        _extra_decoder_kwargs = extra_decoder_kwargs or {}
        self.decoder = DecoderSCVI(
            n_input_decoder,
            n_input,
            n_cat_list=cat_list,
            n_layers=n_layers,
            n_hidden=n_hidden,
            inject_covariates=deeply_inject_covariates,
            use_batch_norm=use_batch_norm_decoder,
            use_layer_norm=use_layer_norm_decoder,
            scale_activation="softplus" if use_size_factor_key else "softmax",
            **_extra_decoder_kwargs,
        )

        # make the library size adaptively
        self.adaptive_library = adaptive_library
        self.library_decoder = lambda x: x
        if encode_covariates and adaptive_library:
            assert n_continuous_cov > 0, "The number of continuous covariates should be greater than 0" + \
                "so the model can adaptively decode the library size"
            n_input_library_decoder = 1 + n_continuous_cov
            self.library_decoder = torch.nn.Sequential(
                torch.nn.Linear(n_input_library_decoder,
                                n_input_library_decoder // 2), torch.nn.GELU(),
                torch.nn.Dropout(dropout_rate),
                torch.nn.Linear(n_input_library_decoder // 2, 1))
        else:
            self.adaptive_library = False

    def _get_inference_input(
        self,
        tensors: Dict[str, torch.Tensor | None],
    ) -> Dict[str, torch.Tensor | None]:
        """Get input tensors for the inference process."""

        return {
            'gene_counts': tensors['gene_counts'],
            'batch_ids': tensors['batch_ids'],
            'cont_covs': tensors.get('cont_covs', None),
            'cat_covs': tensors.get('cat_covs', None),
        }

    def _get_generative_input(
        self,
        tensors: Dict[str, torch.Tensor],
        inference_outputs: Dict[str, torch.Tensor | Distribution | None],
    ) -> Dict[str, torch.Tensor | None]:
        """Get input tensors for the generative process."""
        size_factor = tensors.get('size_factor', None)
        if size_factor is not None:
            size_factor = torch.log(size_factor)

        return {
            'z': inference_outputs['z'],
            'library': inference_outputs['library'],
            'batch_ids': tensors['batch_ids'],
            'y': tensors['y'],
            'cont_covs': tensors.get('cont_covs', None),
            'cat_covs': tensors.get('cat_covs', None),
            'size_factor': size_factor,
        }

    def _compute_local_library_params(
        self,
        batch_index: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Computes local library parameters.

        Compute two tensors of shape (batch_index.shape[0], 1) where each
        element corresponds to the mean and variances, respectively, of the
        log library sizes in the batch the cell corresponds to.
        """
        from torch.nn.functional import linear

        n_batch = self.library_log_means.shape[1]
        local_library_log_means = linear(
            one_hot(batch_index.squeeze(-1), n_batch).float(),
            self.library_log_means)

        local_library_log_vars = linear(
            one_hot(batch_index.squeeze(-1), n_batch).float(),
            self.library_log_vars)

        return local_library_log_means, local_library_log_vars

    def inference(
        self,
        gene_counts: torch.Tensor,
        batch_ids: torch.Tensor,
        cont_covs: torch.Tensor | None = None,
        cat_covs: torch.Tensor | None = None,
        n_samples: int = 1,
    ) -> Dict[str, torch.Tensor | Distribution | None]:
        """Run the regular inference process."""
        gene_counts_ = gene_counts
        if self.use_observed_lib_size:
            library = torch.log(gene_counts.sum(1)).unsqueeze(1)
        if self.log_variational:
            gene_counts_ = torch.log1p(gene_counts_)

        if cont_covs is not None and self.encode_covariates:
            encoder_input = torch.cat((gene_counts_, cont_covs), dim=-1)
        else:
            encoder_input = gene_counts_
        if cat_covs is not None and self.encode_covariates:
            categorical_input = torch.split(cat_covs, 1, dim=1)
        else:
            categorical_input = ()

        if self.batch_representation == "embedding" and self.encode_covariates:
            batch_rep = self.compute_embedding('batch_ids', batch_ids)
            encoder_input = torch.cat([encoder_input, batch_rep], dim=-1)
            qz, z = self.z_encoder(encoder_input, *categorical_input)
        else:
            qz, z = self.z_encoder(encoder_input, batch_ids, *categorical_input)

        ql = None
        if not self.use_observed_lib_size:
            if self.batch_representation == "embedding":
                ql, library_encoded = self.l_encoder(encoder_input,
                                                     *categorical_input)
            else:
                ql, library_encoded = self.l_encoder(encoder_input, batch_ids,
                                                     *categorical_input)
            library = library_encoded

        if n_samples > 1:
            untran_z = qz.sample((n_samples,))
            z = self.z_encoder.z_transformation(untran_z)
            if self.use_observed_lib_size:
                library = library.unsqueeze(0).expand(
                    (n_samples, library.size(0), library.size(1)))
            else:
                library = ql.sample((n_samples,))

        return {
            'z':
                z,  # the latent variable z sampled from the posterior distribution
            'q_z': qz,  # the posterior distribution of the latent variable z
            'q_library': ql,  # the posterior distribution of the library size
            'library':
                library,  # the library size sampled from the posterior distribution
        }

    def generative(
        self,
        z: torch.Tensor,
        library: torch.Tensor,
        batch_ids: torch.Tensor,
        cont_covs: torch.Tensor | None = None,
        cat_covs: torch.Tensor | None = None,
        size_factor: torch.Tensor | None = None,
        y: torch.Tensor | None = None,
        transform_batch: torch.Tensor | None = None,
    ) -> Dict[str, Distribution | None]:
        """Run the generative process."""
        from torch.nn.functional import linear

        from scvi.distributions import NegativeBinomial, Poisson, ZeroInflatedNegativeBinomial

        # TODO: refactor forward function to not rely on y
        # Likelihood distribution
        if cont_covs is None:
            decoder_input = z
        elif z.dim() != cont_covs.dim():
            decoder_input = torch.cat(
                [z, cont_covs.unsqueeze(0).expand(z.size(0), -1, -1)], dim=-1)
        else:
            # for generative process, cont_covs is concatenated to z
            decoder_input = torch.cat([z, cont_covs], dim=-1)

        if cat_covs is not None:
            categorical_input = torch.split(cat_covs, 1, dim=1)
        else:
            categorical_input = ()

        if transform_batch is not None:
            batch_ids = torch.ones_like(batch_ids) * transform_batch

        if not self.use_size_factor_key:
            # MODIFIED: adjust the size_factor based on the conditional information adaptively
            size_factor = library
            if self.adaptive_library and cont_covs is not None:
                size_factor = self.library_decoder(
                    torch.cat([size_factor, cont_covs], dim=-1))

        if self.batch_representation == "embedding":
            batch_rep = self.compute_embedding('batch_ids', batch_ids)
            decoder_input = torch.cat([decoder_input, batch_rep], dim=-1)
            px_scale, px_r, px_rate, px_dropout = self.decoder(
                self.dispersion,
                decoder_input,
                size_factor,
                *categorical_input,
                y,
            )
        else:
            px_scale, px_r, px_rate, px_dropout = self.decoder(
                self.dispersion,
                decoder_input,
                size_factor,
                batch_ids,
                *categorical_input,
                y,
            )

        if self.dispersion == "gene-label":
            px_r = linear(
                one_hot(y.squeeze(-1), self.n_labels).float(),
                self.px_r)  # px_r gets transposed - last dimension is nb genes
        elif self.dispersion == "gene-batch":
            px_r = linear(
                one_hot(batch_ids.squeeze(-1), self.n_batch).float(), self.px_r)
        elif self.dispersion == "gene":
            px_r = self.px_r

        px_r = torch.exp(px_r)

        if self.gene_likelihood == "zinb":
            px = ZeroInflatedNegativeBinomial(
                mu=px_rate,
                theta=px_r,
                zi_logits=px_dropout,
                scale=px_scale,
            )
        elif self.gene_likelihood == "nb":
            px = NegativeBinomial(mu=px_rate, theta=px_r, scale=px_scale)
        elif self.gene_likelihood == "poisson":
            px = Poisson(px_rate, scale=px_scale)

        # Priors
        pl = self.get_prior_library_distribution(batch_ids)
        pz = Normal(torch.zeros_like(z), torch.ones_like(z))

        return {
            'p_gene':
                px,  # the distribution of the gene expression for each cell 
            'p_library':
                pl,  # the distribution of the library size for each cell
            'p_z': pz,  # the prior distribution of the latent variable z
        }

    def get_prior_library_distribution(self,
                                       batch_ids: torch.Tensor) -> Distribution:
        """Get the prior distribution of the library size."""
        if self.use_observed_lib_size:
            return None
        else:
            (
                local_library_log_means,
                local_library_log_vars,
            ) = self._compute_local_library_params(batch_ids)

            return Normal(local_library_log_means,
                          local_library_log_vars.sqrt())

    def loss(
        self,
        tensors: Dict[str, torch.Tensor],
        inference_outputs: Dict[str, torch.Tensor | Distribution | None],
        generative_outputs: Dict[str, Distribution | None],
        kl_weight: float = 3.0,
    ) -> LossOutput:
        """Compute the loss."""
        from torch.distributions import kl_divergence

        x = tensors['gene_counts']
        kl_divergence_z = kl_divergence(inference_outputs['q_z'],
                                        generative_outputs['p_z']).sum(dim=-1)
        if not self.use_observed_lib_size:
            kl_divergence_l = kl_divergence(
                inference_outputs['q_library'],
                generative_outputs['p_library']).sum(dim=1)
        else:
            kl_divergence_l = torch.tensor(0.0, device=x.device)

        reconst_loss = -generative_outputs['p_gene'].log_prob(x).sum(-1)

        kl_local_for_warmup = kl_divergence_z
        kl_local_no_warmup = kl_divergence_l

        weighted_kl_local = kl_weight * kl_local_for_warmup + kl_local_no_warmup

        loss = torch.mean(reconst_loss + weighted_kl_local)

        return LossOutput(
            loss=loss,
            reconstruction_loss=reconst_loss,
            kl_local={
                'kl_library': kl_divergence_l,
                'kl_z': kl_divergence_z,
            },
        )



# -*- coding: utf-8 -*-

import os
import sys

import unittest
import shutil
import numpy as np

# temporary solution for relative imports in case TDC is not installed
# if TDC is installed, no need to use the following line
sys.path.append(
    os.path.abspath(os.path.join(os.path.dirname(__file__), "../..")))
# TODO: add verification for the generation other than simple integration

from tdc_ml.model_server.tokenizers.geneformer import GeneformerTokenizer


def quant_layers(model):
    layer_nums = []
    for name, parameter in model.named_parameters():
        if "layer" in name:
            layer_nums += [int(name.split("layer.")[1].split(".")[0])]
    return int(max(layer_nums)) + 1


class TestModelServer(unittest.TestCase):

    def setUp(self):
        print(os.getcwd())

    def testscGPT(self):
        from tdc_ml.multi_pred.anndata_dataset import DataLoader
        from tdc_ml import tdc_hf_interface
        from tdc_ml.model_server.tokenizers.scgpt import scGPTTokenizer
        import torch
        adata = DataLoader("cellxgene_sample_small",
                           "./data",
                           dataset_names=["cellxgene_sample_small"],
                           no_convert=True).adata
        scgpt = tdc_hf_interface("scGPT")
        model = scgpt.load()  # this line can cause segmentation fault
        tokenizer = scGPTTokenizer()
        gene_ids = adata.var["feature_name"].to_numpy(
        )  # Convert to numpy array
        tokenized_data = tokenizer.tokenize_cell_vectors(
            adata.X.toarray(), gene_ids)
        mask = torch.tensor([x != 0 for x in tokenized_data[0][1]],
                            dtype=torch.bool)
        assert sum(mask) != 0, "FAILURE: mask is empty"
        first_embed = model(tokenized_data[0][0],
                            tokenized_data[0][1],
                            attention_mask=mask)
        print(f"scgpt ran successfully. here is an output {first_embed}")

    def testGeneformerPerturb(self):
        from tdc_ml.multi_pred.perturboutcome import PerturbOutcome
        dataset = "scperturb_drug_AissaBenevolenskaya2021"
        data = PerturbOutcome(dataset)
        adata = data.adata
        tokenizer = GeneformerTokenizer(max_input_size=3)
        adata.var["feature_id"] = adata.var.index.map(
            lambda x: tokenizer.gene_name_id_dict.get(x, 0))
        x = tokenizer.tokenize_cell_vectors(adata,
                                            ensembl_id="feature_id",
                                            ncounts="ncounts")
        cells, _ = x
        assert cells, "FAILURE: cells false-like. Value is = {}".format(cells)
        assert len(cells) > 0, "FAILURE: length of cells <= 0 {}".format(cells)
        from tdc_ml import tdc_hf_interface
        import torch
        geneformer = tdc_hf_interface("Geneformer")
        model = geneformer.load()
        mdim = max(len(cell) for b in cells for cell in b)
        batch = cells[0]
        for idx, cell in enumerate(batch):
            if len(cell) < mdim:
                for _ in range(mdim - len(cell)):
                    cell = np.append(cell, 0)
                batch[idx] = cell
        input_tensor = torch.tensor(batch)
        assert input_tensor.shape[0] == 512, "unexpected batch size"
        assert input_tensor.shape[1] == mdim, f"unexpected gene length {mdim}"
        attention_mask = torch.tensor([[t != 0 for t in cell] for cell in batch
                                      ])
        assert input_tensor.shape[0] == attention_mask.shape[0]
        assert input_tensor.shape[1] == attention_mask.shape[1]
        try:
            outputs = model(input_tensor,
                            attention_mask=attention_mask,
                            output_hidden_states=True)
        except Exception as e:
            raise Exception(
                f"sizes: {input_tensor.shape[0]}, {input_tensor.shape[1]}\n {e}"
            )
        num_out_in_batch = len(outputs.hidden_states[-1])
        input_batch_size = input_tensor.shape[0]
        num_gene_out_in_batch = len(outputs.hidden_states[-1][0])
        assert num_out_in_batch == input_batch_size, f"FAILURE: length doesn't match batch size {num_out_in_batch} vs {input_batch_size}"
        assert num_gene_out_in_batch == mdim, f"FAILURE: out length {num_gene_out_in_batch} doesn't match gene length {mdim}"

    def testscVI(self):
        from tdc_ml.multi_pred.anndata_dataset import DataLoader
        from tdc_ml import tdc_hf_interface

        adata = DataLoader("scvi_test_dataset",
                           "./data",
                           dataset_names=["scvi_test_dataset"],
                           no_convert=True).adata

        scvi = tdc_hf_interface("scVI")
        model = scvi.load()
        output = model(adata)
        print(f"scVI ran successfully. here is an ouput: {output}")

    def testVCGPT(self):
        from tdc_ml.model_server import load_from_hf
        vcgpt = load_from_hf("VCGPT")
        model = vcgpt.load()
        print(f"VCGPT loaded successfully")
        # data
        from tdc_ml.multi_pred.anndata_dataset import DataLoader
        from tdc_ml.utils.load import download_wrapper, pd_load
        adata = DataLoader("vcgpt_sensitivity_demo",
                           "./data",
                           dataset_names=["vcgpt_sensitivity_demo"],
                           no_convert=True).adata
        download_wrapper("vcgpt_gene_vocab", "./data", "vcgpt_gene_vocab")
        gene_vocab = pd_load("vcgpt_gene_vocab", "./data")
        # process data
        from tdc_ml.model_server.model_loaders.instructcell import unify_gene_features
        adata = unify_gene_features(
            adata,
            gene_features=gene_vocab,
            force_gene_symbol_uppercase=False,
        )
        # inference
        # Select a random single-cell sample and extract its gene counts and metadata
        k = np.random.randint(0, len(adata))
        gene_counts = adata[k, :].X.toarray()
        sc_metadata = adata[k, :].obs.iloc[0].to_dict()

        # Define the model prompt with placeholders for metadata and gene expression profile
        prompt = (
            "Given {sequencing_method}, can you predict the response of the single cell {input} from {species} when exposed to {drug} in {tissue}?"
        )

        output = model.predict(
            prompt,
            gene_counts=gene_counts,
            sc_metadata=sc_metadata,
            do_sample=True,
            top_p=0.95,
            top_k=50,
            max_new_tokens=256,
        )
        val = output["text"]
        assert val is not None and len(val) > 0, "VCGPT output is empty"
        assert val in ["Sensitive", "Resistant",
                       "Holiday"], f"VCGPT output is unexpected: {val}"

    def tearDown(self):
        try:
            print(os.getcwd())
            shutil.rmtree(os.path.join(os.getcwd(), "data"))
        except:
            pass



import argparse
import pandas as pd
import numpy as np
from dataclasses import dataclass
from typing import (
    Iterable,
    List,
    Optional,
    Dict,
    Any,
)
from scipy.sparse import lil_matrix
from collections import defaultdict
import anndata
import json
from importlib.util import find_spec
import os
import warnings


@dataclass(frozen=True)
class Template:
    """A data structure representing input and optional output templates."""

    input: str
    output: Optional[str] = None


def str2bool(v: str) -> bool:
    """Convert a string to a boolean value."""
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", 't', 'y', '1'):
        return True
    elif v.lower() in ("no", "false", 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


def read_templates(
    template_dir_name: Optional[str] = None,
    task_type: Optional[str] = None,
) -> Dict[str, List[Template]]:
    """
    Read templates from a directory, structured by task type, and return a dict of template sets.

    The function looks for subdirectories within the given directory, each representing a different task type.
    Each subdirectory should contain a 'templates.json' file that stores template data with input instructions 
    and corresponding output responses. The templates are parsed and returned as a dictionary where the keys 
    are the task types and the values are lists of Template objects. If no directory is provided, the current 
    working directory is used. The function also allows filtering by a specific task type, so only that 
    subdirectory's templates are processed.

    Parameters
    ----------
    template_dir_name: str, optional, default None
        The name of the directory containing task subdirectories with 'templates.json' files. If not provided,
        the current directory is used by default.
    task_type: str, optional, default None 
        The specific task type to filter for. If provided, only the templates from the corresponding subdirectory 
        are read.

    Returns
    -------
    template_set: dict of str to list of Template
        A dict where keys are task types (subdirectory names) and values are lists of Template objects 
        containing input instructions and optional output responses.
    """
    if template_dir_name is None:
        warnings.warn(
            "No template directory is provided. Use the current directory.",
            UserWarning)
        template_dir_name = '.'
    template_set = defaultdict(list)
    have_invalid_template = False

    for sub_dir in os.listdir(template_dir_name):
        if task_type is not None and sub_dir != task_type:
            continue
        path = os.path.join(template_dir_name, sub_dir)
        if os.path.isdir(path):
            template_file_name = os.path.join(path, "templates.json")
            if os.path.exists(template_file_name):
                with open(template_file_name, 'r') as f:
                    templates = json.load(f)
                for template in templates:
                    if "instruction" not in template or "response" not in template:
                        have_invalid_template = True
                    input = template.get("instruction", "{input}")
                    output = template.get("response", "{output}")
                    template_set[sub_dir].append(Template(input, output=output))

    if have_invalid_template:
        warnings.warn(
            "Some templates are invalid. Please ensure that all templates have 'instruction' and 'response' keys. "
            "Those invalid templates will be replaced with the default template, namely '{input}' and '{output}'.",
            UserWarning)
    return template_set


def find_duplicates_uppercase(inputs: Iterable[str]) -> List[bool]:
    """Find duplicates in the input list after converting all elements to uppercase."""
    visited = set()
    res = []

    for x in inputs:
        x_ = x.upper()
        if x_ in visited:
            res.append(False)
        else:
            res.append(True)
            visited.add(x_)

    return res


def unify_gene_features(
    adata: anndata.AnnData,
    gene_features: Iterable[str],
    force_gene_symbol_uppercase: bool = True,
) -> anndata.AnnData:
    """
    Unify the gene features of the AnnData object to match the provided gene features.

    This function adjusts the gene features in an AnnData object, ensuring that only the specified 
    gene features remain. Gene features present in the list but not in the AnnData object will be 
    added with empty data. Optionally, gene symbols can be converted to uppercase. 

    Parameters
    ----------
    adata: anndata.AnnData
        The AnnData object whose gene features need to be unified.
    gene_features: a sequence of str 
        A sequence of genes to unify within the AnnData object.
    force_gene_symbol_uppercase: bool, default True
        If True, both the gene symbols in the AnnData object and the input gene features will be 
        converted to uppercase. If False, case sensitivity will be maintained.

    Returns
    -------
    new_adata: anndata.AnnData
        A new AnnData object with unified gene features.

    Notes
    -----
    - Duplicates may arise if case conversion leads to identical gene symbols. In such cases, 
      a ValueError will be raised unless ``force_gene_symbol_uppercase`` is set to False.
    - Metadata in the AnnData object will not be altered, and the function does not operate 
      in-place, creating a new object instead.

    Examples
    --------
    >>> adata = anndata.AnnData(
    ...     X=np.array([[1, 2], [3, 4]]), 
    ...     var=pd.DataFrame(index=["GeneA", "GeneB"]), 
    ...     obs=pd.DataFrame({"state": ["normal", "cancer"]}, index=["CellA', 'CellB"]),
    ... )
    >>> gene_features = ["GENEA", "GENEC"]
    >>> new_adata = unify_gene_features(adata, gene_features, force_gene_symbol_uppercase=True)
    >>> new_adata.X.toarray()
    array([[1., 0.],
           [3., 0.]])
    >>> new_adata.var_names 
    Index(['GENEA', 'GENEC'], dtype='object')
    >>> new_adata.obs.to_dict()
    {'state': {'CellA': 'normal', 'CellB': 'cancer'}}
    """
    # drop duplicates
    # TO DO: consider gene copies with different esmbl ids
    if force_gene_symbol_uppercase:
        adata = adata[:, find_duplicates_uppercase(adata.var_names)].copy()
        gene_features = [gene.upper() for gene in gene_features]
        if len(set(gene_features)) != len(gene_features):
            raise ValueError("Duplicates occur when gene symbols are capitalized. " + \
                              "To avert it, please remove duplicates in the gene features or set 'force_gene_symbol_uppercase' to False")

    new_var_names = pd.Index(gene_features)
    # lil_matrix is efficient for constructing sparse matrices incrementally
    new_X = lil_matrix((len(adata), len(new_var_names)))

    # drop genes that are not in the gene vocabulary and order the genes
    var_features = {
        gene.upper() if force_gene_symbol_uppercase else gene: i
        for i, gene in enumerate(adata.var_names)
    }
    gene_features = np.array(gene_features)
    indexer = np.arange(len(gene_features))
    src_indexer = indexer[np.vectorize(
        lambda x: gene_features[x] in var_features)(indexer)]
    tgt_indexer = np.vectorize(lambda x: var_features[x])(
        gene_features[src_indexer])
    new_X[:, src_indexer] = adata.X[:, tgt_indexer]

    # convert the lil_matrix to csr_matrix
    # we don't change the metadata of the AnnData object
    return anndata.AnnData(X=new_X.tocsr(),
                           obs=adata.obs,
                           var=pd.DataFrame(index=new_var_names))


def set_global_random_seed(seed: int,
                           libraries: Optional[Iterable[str]] = None) -> None:
    """Set the global random seed for reproducibility."""
    if libraries is None:
        libraries = ["numpy", "torch", "random"]

    for lib in libraries:
        if lib == "numpy":
            if find_spec("numpy") is not None:
                import numpy as np
                np.random.seed(seed)
        elif lib == "torch":
            if find_spec("torch") is not None:
                import torch
                torch.manual_seed(seed)
                torch.cuda.manual_seed(seed)
                torch.cuda.manual_seed_all(seed)
                torch.backends.cudnn.benchmark = False
                torch.backends.cudnn.deterministic = True
        elif lib == "random":
            import random
            random.seed(seed)
        else:
            raise ValueError(
                f"Library {lib} is not supported for setting the global seed.")

    return


def parse_parameters(parameters: Dict[str, Any]) -> Dict[str, Dict[str, Any]]:
    """Parse the model parameters."""
    model_parameters = {}
    for key, value in parameters.items():
        name, sub_key = key.split("::", 1)
        if name not in model_parameters:
            model_parameters[name] = {}
        model_parameters[name][sub_key] = value
    return model_parameters



import torch
import torch.nn as nn
from typing import (
    Literal,
    Iterable,
    Optional,
    Callable,
    List,
    Dict,
)
import collections
from torch.distributions import Normal


def _identity(x):
    return x


def _partial_freeze_hook_factory(
        freeze: int) -> Callable[[torch.Tensor], torch.Tensor]:
    """Factory for a hook that freezes the first ``freeze`` entries in the gradient.

    Parameters
    ----------
    freeze
        Specifies the number of entries to freeze in the gradient.
    """

    def _partial_freeze_hook(grad: torch.Tensor) -> torch.Tensor:
        grad = grad.clone()
        grad[:freeze] = 0.0
        return grad

    return _partial_freeze_hook


class FCLayers(nn.Module):
    """A helper class to build fully-connected layers for a neural network.

    Parameters
    ----------
    n_in
        The dimensionality of the input
    n_out
        The dimensionality of the output
    n_cat_list
        A list containing, for each category of interest,
        the number of categories. Each category will be
        included using a one-hot encoding.
    n_layers
        The number of fully-connected hidden layers
    n_hidden
        The number of nodes per hidden layer
    dropout_rate
        Dropout rate to apply to each of the hidden layers
    use_batch_norm
        Whether to have `BatchNorm` layers or not
    use_layer_norm
        Whether to have `LayerNorm` layers or not
    use_activation
        Whether to have layer activation or not
    bias
        Whether to learn bias in linear layers or not
    inject_covariates
        Whether to inject covariates in each layer, or just the first (default).
    activation_fn
        Which activation function to use
    """

    def __init__(
            self,
            n_in: int,
            n_out: int,
            n_cat_list: Iterable[int] = None,
            n_layers: int = 1,
            n_hidden: int = 128,
            dropout_rate: float = 0.1,
            use_batch_norm: bool = True,
            use_layer_norm: bool = False,
            use_activation: bool = True,
            bias: bool = True,
            inject_covariates: bool = True,
            activation_fn: nn.Module = nn.GELU,  # considering transformer 
    ):
        super().__init__()
        self.inject_covariates = inject_covariates
        layers_dim = [n_in] + (n_layers - 1) * [n_hidden] + [n_out]

        if n_cat_list is not None:
            # n_cat = 1 will be ignored
            self.n_cat_list = [
                n_cat if n_cat > 1 else 0 for n_cat in n_cat_list
            ]
        else:
            self.n_cat_list = []

        cat_dim = sum(self.n_cat_list)
        self.fc_layers = nn.Sequential(
            collections.OrderedDict([
                (
                    f"Layer {i}",
                    nn.Sequential(
                        nn.Linear(
                            n_in + cat_dim * self.inject_into_layer(i),
                            n_out,
                            bias=bias,
                        ),
                        # non-default params come from defaults in original Tensorflow
                        # implementation
                        nn.BatchNorm1d(n_out, momentum=0.01, eps=0.001)
                        if use_batch_norm else None,
                        nn.LayerNorm(n_out, elementwise_affine=False)
                        if use_layer_norm else None,
                        activation_fn() if use_activation else None,
                        nn.Dropout(
                            p=dropout_rate) if dropout_rate > 0 else None,
                    ),
                ) for i, (
                    n_in,
                    n_out) in enumerate(zip(layers_dim[:-1], layers_dim[1:]))
            ]))

    def inject_into_layer(self, layer_num) -> bool:
        """Helper to determine if covariates should be injected."""
        user_cond = layer_num == 0 or (layer_num > 0 and self.inject_covariates)
        return user_cond

    def set_online_update_hooks(self, hook_first_layer=True):
        """Set online update hooks."""
        self.hooks = []

        def _hook_fn_weight(grad):
            categorical_dims = sum(self.n_cat_list)
            new_grad = torch.zeros_like(grad)
            if categorical_dims > 0:
                new_grad[:, -categorical_dims:] = grad[:, -categorical_dims:]
            return new_grad

        def _hook_fn_zero_out(grad):
            return grad * 0

        for i, layers in enumerate(self.fc_layers):
            for layer in layers:
                if i == 0 and not hook_first_layer:
                    continue
                if isinstance(layer, nn.Linear):
                    if self.inject_into_layer(i):
                        w = layer.weight.register_hook(_hook_fn_weight)
                    else:
                        w = layer.weight.register_hook(_hook_fn_zero_out)
                    self.hooks.append(w)
                    b = layer.bias.register_hook(_hook_fn_zero_out)
                    self.hooks.append(b)

    def forward(
        self,
        x: torch.Tensor,
        *cat_list: int,
        output_hidden_states: bool = False,
    ):
        """Forward computation on ``x``.

        Parameters
        ----------
        x
            tensor of values with shape ``(n_in,)``
        cat_list
            list of category membership(s) for this sample
        output_hidden_states
            If ``True``, return the hidden states of the network

        Returns
        -------
        :class:`torch.Tensor`
            tensor of shape ``(n_out,)`` or a tuple of tensors if ``output_hidden_states`` is ``True``
        """
        one_hot_cat_list = [
        ]  # for generality in this list many indices useless.

        if len(self.n_cat_list) > len(cat_list):
            raise ValueError(
                "nb. categorical args provided doesn't match init. params.")
        for n_cat, cat in zip(self.n_cat_list, cat_list):
            if n_cat and cat is None:
                raise ValueError(
                    "cat not provided while n_cat != 0 in init. params.")
            if n_cat > 1:  # n_cat = 1 will be ignored - no additional information
                if cat.size(1) != n_cat:
                    one_hot_cat = nn.functional.one_hot(cat.squeeze(-1), n_cat)
                else:
                    one_hot_cat = cat  # cat has already been one_hot encoded
                one_hot_cat_list += [one_hot_cat]

        hidden_states = {}
        for i, layers in enumerate(self.fc_layers):
            for layer in layers:
                if layer is not None:
                    if isinstance(layer, nn.BatchNorm1d):
                        if x.dim() == 3:
                            x = torch.cat([
                                (layer(slice_x)).unsqueeze(0) for slice_x in x
                            ],
                                          dim=0)
                        else:
                            x = layer(x)
                    else:
                        if isinstance(layer,
                                      nn.Linear) and self.inject_into_layer(i):
                            if x.dim() == 3:
                                one_hot_cat_list_layer = [
                                    o.unsqueeze(0).expand(
                                        (x.size(0), o.size(0), o.size(1)))
                                    for o in one_hot_cat_list
                                ]
                            else:
                                one_hot_cat_list_layer = one_hot_cat_list
                            x = torch.cat((x, *one_hot_cat_list_layer), dim=-1)
                        x = layer(x)
            if output_hidden_states:
                hidden_states[f"Layer {i}"] = x

        return x if not output_hidden_states else (x, hidden_states)


class Encoder(nn.Module):
    """Encode data of ``n_input`` dimensions into a latent space of ``n_output`` dimensions.

    Uses a fully-connected neural network of ``n_hidden`` layers.

    Parameters
    ----------
    n_input
        The dimensionality of the input (data space)
    n_output
        The dimensionality of the output (latent space)
    n_cat_list
        A list containing the number of categories
        for each category of interest. Each category will be
        included using a one-hot encoding
    n_layers
        The number of fully-connected hidden layers
    n_hidden
        The number of nodes per hidden layer
    dropout_rate
        Dropout rate to apply to each of the hidden layers
    distribution
        Distribution of z
    var_eps
        Minimum value for the variance;
        used for numerical stability
    var_activation
        Callable used to ensure positivity of the variance.
        Defaults to :meth:`torch.exp`.
    return_dist
        Return directly the distribution of z instead of its parameters.
    **kwargs
        Keyword args for :class:`~scvi.nn.FCLayers`
    """

    def __init__(
        self,
        n_input: int,
        n_output: int,
        n_cat_list: Iterable[int] = None,
        n_layers: int = 1,
        n_hidden: int = 128,
        dropout_rate: float = 0.1,
        distribution: str = "normal",
        var_eps: float = 1e-4,
        var_activation: Optional[Callable] = None,
        return_dist: bool = False,
        **kwargs,
    ):
        super().__init__()

        self.distribution = distribution
        self.var_eps = var_eps
        self.encoder = FCLayers(
            n_in=n_input,
            n_out=n_hidden,
            n_cat_list=n_cat_list,
            n_layers=n_layers,
            n_hidden=n_hidden,
            dropout_rate=dropout_rate,
            **kwargs,
        )
        self.mean_encoder = nn.Linear(n_hidden, n_output)
        self.var_encoder = nn.Linear(n_hidden, n_output)
        self.return_dist = return_dist

        if distribution == "ln":
            self.z_transformation = nn.Softmax(dim=-1)
        else:
            self.z_transformation = _identity
        self.var_activation = torch.exp if var_activation is None else var_activation

    def forward(self, x: torch.Tensor, *cat_list: int):
        r"""The forward computation for a single sample.

         #. Encodes the data into latent space using the encoder network
         #. Generates a mean \\( q_m \\) and variance \\( q_v \\)
         #. Samples a new value from an i.i.d. multivariate normal
            \\( \\sim Ne(q_m, \\mathbf{I}q_v) \\)

        Parameters
        ----------
        x
            tensor with shape (n_input,)
        cat_list
            list of category membership(s) for this sample

        Returns
        -------
        3-tuple of :py:class:`torch.Tensor`
            tensors of shape ``(n_latent,)`` for mean and var, and sample

        """
        # Parameters for latent distribution
        q = self.encoder(x, *cat_list)
        q_m = self.mean_encoder(q)
        q_v = self.var_activation(self.var_encoder(q)) + self.var_eps
        dist = Normal(q_m, q_v.sqrt())
        latent = self.z_transformation(dist.rsample())
        if self.return_dist:
            return dist, latent
        return q_m, q_v, latent


class DecoderSCVI(nn.Module):
    """Decodes data from latent space of ``n_input`` dimensions into ``n_output`` dimensions.

    Uses a fully-connected neural network of ``n_hidden`` layers.

    Parameters
    ----------
    n_input
        The dimensionality of the input (latent space)
    n_output
        The dimensionality of the output (data space)
    n_cat_list
        A list containing the number of categories
        for each category of interest. Each category will be
        included using a one-hot encoding
    n_layers
        The number of fully-connected hidden layers
    n_hidden
        The number of nodes per hidden layer
    dropout_rate
        Dropout rate to apply to each of the hidden layers
    inject_covariates
        Whether to inject covariates in each layer, or just the first (default).
    use_batch_norm
        Whether to use batch norm in layers
    use_layer_norm
        Whether to use layer norm in layers
    scale_activation
        Activation layer to use for px_scale_decoder
    **kwargs
        Keyword args for :class:`~scvi.nn.FCLayers`.
    """

    def __init__(
        self,
        n_input: int,
        n_output: int,
        n_cat_list: Iterable[int] = None,
        n_layers: int = 1,
        n_hidden: int = 128,
        inject_covariates: bool = True,
        use_batch_norm: bool = False,
        use_layer_norm: bool = False,
        scale_activation: Literal["softmax", "softplus"] = "softmax",
        **kwargs,
    ):
        super().__init__()
        self.px_decoder = FCLayers(
            n_in=n_input,
            n_out=n_hidden,
            n_cat_list=n_cat_list,
            n_layers=n_layers,
            n_hidden=n_hidden,
            dropout_rate=0,
            inject_covariates=inject_covariates,
            use_batch_norm=use_batch_norm,
            use_layer_norm=use_layer_norm,
            **kwargs,
        )

        # mean gamma
        if scale_activation == "softmax":
            px_scale_activation = nn.Softmax(dim=-1)
        elif scale_activation == "softplus":
            px_scale_activation = nn.Softplus()
        self.px_scale_decoder = nn.Sequential(
            nn.Linear(n_hidden, n_output),
            px_scale_activation,
        )

        # dispersion: here we only deal with gene-cell dispersion case
        self.px_r_decoder = nn.Linear(n_hidden, n_output)

        # dropout
        self.px_dropout_decoder = nn.Linear(n_hidden, n_output)

    def forward(
        self,
        dispersion: str,
        z: torch.Tensor,
        library: torch.Tensor,
        *cat_list: int,
    ):
        """The forward computation for a single sample.

         #. Decodes the data from the latent space using the decoder network
         #. Returns parameters for the ZINB distribution of expression
         #. If ``dispersion != 'gene-cell'`` then value for that param will be ``None``

        Parameters
        ----------
        dispersion
            One of the following

            * ``'gene'`` - dispersion parameter of NB is constant per gene across cells
            * ``'gene-batch'`` - dispersion can differ between different batches
            * ``'gene-label'`` - dispersion can differ between different labels
            * ``'gene-cell'`` - dispersion can differ for every gene in every cell
        z :
            tensor with shape ``(n_input,)``
        library_size
            library size
        cat_list
            list of category membership(s) for this sample

        Returns
        -------
        4-tuple of :py:class:`torch.Tensor`
            parameters for the ZINB distribution of expression

        """
        # The decoder returns values for the parameters of the ZINB distribution
        px = self.px_decoder(z, *cat_list)
        px_scale = self.px_scale_decoder(px)
        px_dropout = self.px_dropout_decoder(px)
        # Clamp to high value: exp(12) ~ 160000 to avoid nans (computational stability)
        px_rate = torch.exp(library) * px_scale  # torch.clamp( , max=12)
        px_r = self.px_r_decoder(px) if dispersion == "gene-cell" else None
        return px_scale, px_r, px_rate, px_dropout


class Embedding(nn.Embedding):
    """``EXPERIMENTAL`` Embedding layer with utility methods for extending.

    Notes
    -----
    Lifecycle: experimental in v1.2.
    """

    @classmethod
    def extend(
        cls,
        embedding: "Embedding",
        init: int | List[int],
        freeze_prev: bool = True,
    ) -> "Embedding":
        """Factory class method for extending an existing :class:`~scvi.nn.Embedding`.

        Initializes new embeddings with random values or copies from the original embedding.

        Parameters
        ----------
        embedding
            Embedding layer to extend. Not modified in-place as a new instance is returned.
        init
            If an ``int``, specifies the number of new random embeddings to initialize with
            :func:`~torch.nn.init.normal_`. If a ``list[int]``, initializes ``len(init)`` new
            embeddings with the values of the given indices in the original embedding.
        freeze_prev
            If ``True``, gradients for the original embeddings are set to zero.

        Returns
        -------
        New :class:`~scvi.nn.Embedding` instance with :attr:`~scvi.nn.Embedding.num_embeddings`
        increased by the specified ``init`` and with gradients frozen for the first
        ``embedding.num_embeddings`` entries if ``freeze_prev`` is ``True``.
        """
        if isinstance(init, int):
            if init <= 0:
                raise ValueError(f"`init` must be greater than 0, got {init}")
            weight = torch.empty((init, embedding.embedding_dim),
                                 device=embedding.weight.device)
            nn.init.normal_(weight)
        elif isinstance(init, list):
            weight = embedding.weight[init].clone()
        else:
            raise TypeError(
                f"`init` must be an `int` or a `list[int]`, got {type(init)}")

        weight = torch.cat([embedding.weight.clone(), weight], dim=0)
        new_embedding = cls(
            num_embeddings=weight.shape[0],
            embedding_dim=weight.shape[1],
            _weight=weight,
            padding_idx=embedding.padding_idx,
            max_norm=embedding.max_norm,
            norm_type=embedding.norm_type,
            scale_grad_by_freq=embedding.scale_grad_by_freq,
            sparse=embedding.sparse,
        )
        if freeze_prev:
            new_embedding.weight.register_hook(
                _partial_freeze_hook_factory(embedding.num_embeddings))

        return new_embedding

    def _load_from_state_dict(self, state_dict: Dict[str, torch.Tensor], *args,
                              **kwargs) -> None:
        """Load from a state dict. Overrides initialization parameters with the state dict.

        This is necessary because model constructors will pass in the original parameters, which
        will not match the state dict if the embedding was extended. This method overrides the
        correct attributes based on the state dict.
        """
        key = [key for key in state_dict.keys() if "weight" in key][0]
        self.weight = nn.Parameter(state_dict[key])
        self.num_embeddings, self.embedding_dim = self.weight.shape[
            0], self.weight.shape[1]
        return super()._load_from_state_dict(state_dict, *args, **kwargs)



from transformers import (
    PreTrainedModel,
    PreTrainedTokenizer,
    PreTrainedTokenizerFast,
    PretrainedConfig,
)
from typing import (
    Dict,
    Tuple,
    List,
    Optional,
)
import torch.nn as nn
import warnings


def _check_modality_name(
        tokenizer: PreTrainedTokenizer | PreTrainedTokenizerFast,
        special_tokens_dict: Dict[str, str | List[str]]) -> None:
    """Check if the special tokens are already in the vocabulary and if they are duplicated."""
    vocab = tokenizer.get_vocab()
    visited = set()
    for tokens in special_tokens_dict.values():
        if isinstance(tokens, str):
            tokens = [tokens]
        for token in tokens:
            if token in vocab:
                raise ValueError(f"Token {token} is already in the vocabulary.")
            if token in visited:
                raise ValueError(f"Token {token} is duplicated.")
            visited.add(token)


def _validate_added_special_tokens(tokenizer: PreTrainedTokenizer |
                                   PreTrainedTokenizerFast,
                                   added_tokens: List[str]) -> None:
    """Check if the added tokens are correctly added to the tokenizer."""
    added_tokens = [token for token in set(added_tokens)]
    new_integers = []
    for token in added_tokens:
        int_list = tokenizer.encode(token, add_special_tokens=False)
        # we assume that the special token is not split into multiple tokens
        assert len(
            int_list
        ) == 1, f"Token '{token}' is encoded into multiple integers: {int_list}."
        new_integers.append(int_list[0])
        if tokenizer.decode(int_list) != token:
            warnings.warn(
                f"For '{token}', the reconstructed token is "
                f"'{tokenizer.decode(int_list)}'.", UserWarning)

    added_tokens = [' '.join(added_tokens), "".join(added_tokens)]
    for token in added_tokens:
        int_list = tokenizer.encode(token, add_special_tokens=False)
        # ensure the combination of the added tokens is split as expected
        assert int_list == new_integers, \
            f"It seems that the result ({int_list}) of encoding '{token}' is not correct. {new_integers} is expected."


def prepare_cell_text_llm(
    model: PreTrainedModel | nn.Module,
    tokenizer: PreTrainedTokenizer | PreTrainedTokenizerFast,
    modality_tag: str = "CELL",
    num_signal_tokens: int = 1,
    ignore_index: Optional[int] = None,
    pad_token_id: Optional[int] = None,
    eos_token_id: Optional[int] = None,
    pad_to_multiple_of: Optional[int] = None,
) -> Tuple[PreTrainedModel | nn.Module, PreTrainedTokenizer |
           PreTrainedTokenizerFast]:
    """
    Prepare a model and tokenizer for handling cell-language tasks.

    This function helps in adapting a model and its tokenizer for a new modality, by adding
    special tokens that represent start, end, and signal tokens for that modality, which 
    can be useful in tasks such as processing cell-specific data in large language models (LLMs).

    Parameters
    ----------
    model: transformers.PreTrainedModel or torch.nn.Module
        The model to be adapted. This can be any pre-trained model from the transformers library 
        or a custom model that supports resizing its token embeddings.
    tokenizer: transformers.PreTrainedTokenizer or transformers.PreTrainedTokenizerFast
        The tokenizer to be adapted. This will have special tokens added for the new modality.
    modality_tag: str, default "CELL"
        The tag that represents the modality to be added to the tokenizer. It will be used to create
        start and end tags as well as signal tokens.
    num_signal_tokens: int, default 1
        The number of signal tokens to add. These tokens serve as signals enabling the model to generate
        data in the specified modality.
    ignore_index: int, optional, default None 
        Index to be ignored during loss computation in models like sequence classification or generation.
        If not provided, the model's ``ignore_index`` should be set.
    pad_token_id: int, optional, default None 
        Token ID to be used for padding. If not provided, the model's ``pad_token_id`` should be set.
    eos_token_id: int, optional, default None 
        Token ID to represent the end of the sequence. If not provided, the model's ``eos_token_id`` 
        should be set.
    pad_to_multiple_of: int, optional, default None 
        If set will pad the embedding matrix to a multiple of the provided value after resizing embedding 
        layer. This is especially useful to enable the use of Tensor Cores on NVIDIA hardware with compute 
        capability ``>= 7.5`` (Volta), or on TPUs which benefit from having sequence lengths be a multiple 
        of 128.

    Returns
    -------
    model: transformers.PreTrainedModel or torch.nn.Module
        The modified model with resized token embeddings to accommodate the new special tokens.
    tokenizer: transformers.PreTrainedTokenizer or transformers.PreTrainedTokenizerFast
        The modified tokenizer with the added special tokens for the new modality.

    Notes
    -----
    - This function adds special tokens like ``<MODALITY>``, ``</MODALITY>``, and signal tokens 
      ``<MODALITY1>``, ``<MODALITY2>``, etc., where ``MODALITY`` is derived from the ``modality_tag``.
      The placeholder token ``<MODALITY0>`` is also added to the tokenizer, which is used to identify the 
      position of the modality data within the text data.

    Examples
    --------
    >>> from transformers import AutoTokenizer, AutoModelForSeq2SeqLM
    >>> MODEL_PATH = "t5-base" 
    >>> tokenizer = AutoTokenizer.from_pretrained(MODEL_PATH)
    >>> model = AutoModelForSeq2SeqLM.from_pretrained(MODEL_PATH)
    >>> model, tokenizer = prepare_cell_text_llm(
    ...     model, 
    ...     tokenizer, 
    ...     modality_tag="CELL",
    ...     ignore_index=-100, 
    ...     pad_to_multiple_of=8, 
    ... )
    Following 4 tokens are added for modality CELL:
    ['<CELL>', '</CELL>', '<CELL0>', '<CELL1>']
    >>> tokenizer.get_added_vocab() 
    {'</CELL>': 32101, '<CELL0>': 32102, '<CELL1>': 32103, '<CELL>': 32100}
    >>> model.config.ignore_index
    -100
    """
    if not hasattr(model, "config"):
        model.config = PretrainedConfig()
    if not hasattr(model, "resize_token_embeddings"):
        raise ValueError(
            "The model should support resizing its token embeddings.")
    special_tokens_dict = {}

    # e.g. if there is a modality called "CELL"
    # start tag <CELL>, end tag </CELL>, placeholder <CELL0>
    # and some signal tokens like <CELL1>, <CELL2>, <CELL3> etc are added to the tokenizer
    # note that we make sure that the order of adding the special tokens is fixed
    special_tokens_dict["start_tag"] = f"<{modality_tag}>"
    special_tokens_dict["end_tag"] = f"</{modality_tag}>"
    special_tokens_dict["placeholder"] = f"<{modality_tag}0>"
    special_tokens_dict["signal_tokens"] = [
        f"<{modality_tag}{i}>" for i in range(1, num_signal_tokens + 1)
    ]
    _check_modality_name(tokenizer, special_tokens_dict)

    # we view these tokens as special tokens so they are not
    # affected by the tokenizer's normalization or splitting
    added_tokens = [
        special_tokens_dict["start_tag"], special_tokens_dict["end_tag"],
        special_tokens_dict["placeholder"]
    ] + special_tokens_dict["signal_tokens"]
    # we don't want to replace the existing special tokens
    tokenizer.add_special_tokens(
        {"additional_special_tokens": added_tokens},
        replace_additional_special_tokens=False,
    )
    _validate_added_special_tokens(tokenizer, added_tokens)
    print(
        f"Following {len(added_tokens)} tokens are added for modality {modality_tag}:\n{added_tokens}"
    )

    # TO DO: placeholder tokens should be not added to the model's embedding layer
    model.resize_token_embeddings(len(tokenizer),
                                  pad_to_multiple_of=pad_to_multiple_of)
    model.config.special_tokens_dict = special_tokens_dict
    added_vocab = tokenizer.get_added_vocab()
    special_tokens_index_dict = {}
    for token_type in special_tokens_dict.keys():
        tokens = special_tokens_dict[token_type]
        if isinstance(tokens, str):
            special_tokens_index_dict[token_type] = added_vocab[tokens]
        else:
            special_tokens_index_dict["first_signal_token"] = added_vocab[
                tokens[0]]
    model.config.special_tokens_index_dict = special_tokens_index_dict

    if pad_token_id is not None:
        model.config.pad_token_id = pad_token_id
    else:
        assert hasattr(
            model.config, "pad_token_id"
        ), "Please provide pad_token_id argument or set model.config.pad_token_id"
    if eos_token_id is not None:
        model.config.eos_token_id = eos_token_id
    else:
        assert hasattr(
            model.config, "eos_token_id"
        ), "Please provide eos_token_id argument or set model.config.eos_token_id"
    if ignore_index is not None:
        model.config.ignore_index = ignore_index
    else:
        assert hasattr(
            model.config, "ignore_index"
        ), "Please provide ignore_index argument or set model.config.ignore_index"
    model.config.num_signal_tokens = num_signal_tokens

    return model, tokenizer
